@@ -1,17 +1,9 @@
 /*
- * Copyright (C) 2009-2015  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2009-2016  Internet Systems Consortium, Inc. ("ISC")
  *
- * Permission to use, copy, modify, and/or distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
- *
- * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
- * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
- * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
- * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
- * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
- * PERFORMANCE OF THIS SOFTWARE.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
 #include <config.h>
@@ -131,6 +123,7 @@ typedef struct resctx {
 	isc_boolean_t		want_dnssec;
 	isc_boolean_t		want_validation;
 	isc_boolean_t		want_cdflag;
+	isc_boolean_t		want_tcp;
 
 	/* Locked */
 	ISC_LINK(struct resctx)	link;
@@ -221,6 +214,7 @@ typedef struct updatectx {
 	unsigned int			magic;
 	isc_mutex_t			lock;
 	dns_client_t			*client;
+	isc_boolean_t			want_tcp;
 
 	/* Locked */
 	dns_request_t			*updatereq;
@@ -783,6 +777,8 @@ start_fetch(resctx_t *rctx) {
 		fopts |= DNS_FETCHOPT_NOCDFLAG;
 	if (!rctx->want_validation)
 		fopts |= DNS_FETCHOPT_NOVALIDATE;
+	if (rctx->want_tcp)
+		fopts |= DNS_FETCHOPT_TCP;
 
 	result = dns_resolver_createfetch(rctx->view->resolver,
 					  dns_fixedname_name(&rctx->name),
@@ -1325,11 +1321,11 @@ dns_client_startresolve(dns_client_t *client, dns_name_t *name,
 	dns_view_t *view = NULL;
 	dns_clientresevent_t *event = NULL;
 	resctx_t *rctx = NULL;
-	isc_task_t *clone = NULL;
+	isc_task_t *tclone = NULL;
 	isc_mem_t *mctx;
 	isc_result_t result;
 	dns_rdataset_t *rdataset, *sigrdataset;
-	isc_boolean_t want_dnssec, want_validation, want_cdflag;
+	isc_boolean_t want_dnssec, want_validation, want_cdflag, want_tcp;
 
 	REQUIRE(DNS_CLIENT_VALID(client));
 	REQUIRE(transp != NULL && *transp == NULL);
@@ -1347,14 +1343,15 @@ dns_client_startresolve(dns_client_t *client, dns_name_t *name,
 	want_dnssec = ISC_TF((options & DNS_CLIENTRESOPT_NODNSSEC) == 0);
 	want_validation = ISC_TF((options & DNS_CLIENTRESOPT_NOVALIDATE) == 0);
 	want_cdflag = ISC_TF((options & DNS_CLIENTRESOPT_NOCDFLAG) == 0);
+	want_tcp = ISC_TF((options & DNS_CLIENTRESOPT_TCP) != 0);
 
 	/*
 	 * Prepare some intermediate resources
 	 */
-	clone = NULL;
-	isc_task_attach(task, &clone);
+	tclone = NULL;
+	isc_task_attach(task, &tclone);
 	event = (dns_clientresevent_t *)
-		isc_event_allocate(mctx, clone, DNS_EVENT_CLIENTRESDONE,
+		isc_event_allocate(mctx, tclone, DNS_EVENT_CLIENTRESDONE,
 				   action, arg, sizeof(*event));
 	if (event == NULL) {
 		result = ISC_R_NOMEMORY;
@@ -1404,6 +1401,7 @@ dns_client_startresolve(dns_client_t *client, dns_name_t *name,
 	rctx->want_dnssec = want_dnssec;
 	rctx->want_validation = want_validation;
 	rctx->want_cdflag = want_cdflag;
+	rctx->want_tcp = want_tcp;
 	ISC_LIST_INIT(rctx->namelist);
 	rctx->event = event;
 
@@ -1429,7 +1427,7 @@ dns_client_startresolve(dns_client_t *client, dns_name_t *name,
 	}
 	if (event != NULL)
 		isc_event_free(ISC_EVENT_PTR(&event));
-	isc_task_detach(&clone);
+	isc_task_detach(&tclone);
 	dns_view_detach(&view);
 
 	return (result);
@@ -1728,12 +1726,11 @@ dns_client_startrequest(dns_client_t *client, dns_message_t *qmessage,
 {
 	isc_result_t result;
 	dns_view_t *view = NULL;
-	isc_task_t *clone = NULL;
+	isc_task_t *tclone = NULL;
 	dns_clientreqevent_t *event = NULL;
 	reqctx_t *ctx = NULL;
 	dns_tsectype_t tsectype = dns_tsectype_none;
-
-	UNUSED(options);
+	unsigned int reqoptions;
 
 	REQUIRE(DNS_CLIENT_VALID(client));
 	REQUIRE(qmessage != NULL);
@@ -1753,10 +1750,14 @@ dns_client_startrequest(dns_client_t *client, dns_message_t *qmessage,
 	if (result != ISC_R_SUCCESS)
 		return (result);
 
-	clone = NULL;
-	isc_task_attach(task, &clone);
+	reqoptions = 0;
+	if ((options & DNS_CLIENTREQOPT_TCP) != 0)
+		reqoptions |= DNS_REQUESTOPT_TCP;
+
+	tclone = NULL;
+	isc_task_attach(task, &tclone);
 	event = (dns_clientreqevent_t *)
-		isc_event_allocate(client->mctx, clone,
+		isc_event_allocate(client->mctx, tclone,
 				   DNS_EVENT_CLIENTREQDONE,
 				   action, arg, sizeof(*event));
 	if (event == NULL) {
@@ -1795,7 +1796,7 @@ dns_client_startrequest(dns_client_t *client, dns_message_t *qmessage,
 
 	ctx->request = NULL;
 	result = dns_request_createvia3(view->requestmgr, qmessage, NULL,
-					server, options, ctx->tsigkey,
+					server, reqoptions, ctx->tsigkey,
 					timeout, udptimeout, udpretries,
 					client->task, request_done, ctx,
 					&ctx->request);
@@ -1815,7 +1816,7 @@ dns_client_startrequest(dns_client_t *client, dns_message_t *qmessage,
 	}
 	if (event != NULL)
 		isc_event_free(ISC_EVENT_PTR(&event));
-	isc_task_detach(&clone);
+	isc_task_detach(&tclone);
 	dns_view_detach(&view);
 
 	return (result);
@@ -1944,7 +1945,7 @@ update_done(isc_task_t *task, isc_event_t *event) {
 	dns_message_t *answer = NULL;
 	updatectx_t *uctx = event->ev_arg;
 	dns_client_t *client;
-	unsigned int timeout;
+	unsigned int timeout, reqoptions;
 
 	UNUSED(task);
 
@@ -1977,18 +1978,35 @@ update_done(isc_task_t *task, isc_event_t *event) {
 	LOCK(&uctx->lock);
 	uctx->currentserver = ISC_LIST_NEXT(uctx->currentserver, link);
 	dns_request_destroy(&uctx->updatereq);
-	if (result != ISC_R_SUCCESS && !uctx->canceled &&
-	    uctx->currentserver != NULL) {
+	/*
+	 * Moving on to the next server shouldn't change the result
+	 * for NXDOMAIN, YXDOMAIN, NXRRSET and YXRRSET as they
+	 * indicate a prerequisite failure.  REFUSED should also
+	 * be consistent across all servers but often isn't as that
+	 * is policy rather that zone content driven (slaves that
+	 * aren't willing to forward should return NOTIMPL).  NOTZONE
+	 * indicates that we stuffed up the request construction so
+	 * don't retry.
+	 */
+	if (result != ISC_R_SUCCESS && result != DNS_R_NXDOMAIN &&
+	    result != DNS_R_YXDOMAIN && result != DNS_R_YXRRSET &&
+	    result != DNS_R_NXRRSET && result != DNS_R_NOTZONE &&
+	    !uctx->canceled && uctx->currentserver != NULL)
+	{
 		dns_message_renderreset(uctx->updatemsg);
 		dns_message_settsigkey(uctx->updatemsg, NULL);
 
 		timeout = client->update_timeout / uctx->nservers;
 		if (timeout < MIN_UPDATE_TIMEOUT)
 			timeout = MIN_UPDATE_TIMEOUT;
+		reqoptions = 0;
+		if (uctx->want_tcp)
+			reqoptions |= DNS_REQUESTOPT_TCP;
 		result = dns_request_createvia3(uctx->view->requestmgr,
 						uctx->updatemsg,
 						NULL,
-						uctx->currentserver, 0,
+						uctx->currentserver,
+						reqoptions,
 						uctx->tsigkey,
 						timeout,
 						client->update_udptimeout,
@@ -2015,7 +2033,7 @@ send_update(updatectx_t *uctx) {
 	dns_name_t *name = NULL;
 	dns_rdataset_t *rdataset = NULL;
 	dns_client_t *client = uctx->client;
-	unsigned int timeout;
+	unsigned int timeout, reqoptions;
 
 	REQUIRE(uctx->zonename != NULL && uctx->currentserver != NULL);
 
@@ -2042,10 +2060,13 @@ send_update(updatectx_t *uctx) {
 	timeout = client->update_timeout / uctx->nservers;
 	if (timeout < MIN_UPDATE_TIMEOUT)
 		timeout = MIN_UPDATE_TIMEOUT;
+	reqoptions = 0;
+	if (uctx->want_tcp)
+		reqoptions |= DNS_REQUESTOPT_TCP;
 	result = dns_request_createvia3(uctx->view->requestmgr,
 					uctx->updatemsg,
-					NULL, uctx->currentserver, 0,
-					uctx->tsigkey, timeout,
+					NULL, uctx->currentserver,
+					reqoptions, uctx->tsigkey, timeout,
 					client->update_udptimeout,
 					client->update_udpretries,
 					client->task, update_done, uctx,
@@ -2182,6 +2203,7 @@ process_soa(updatectx_t *uctx, dns_rdataset_t *soaset, dns_name_t *soaname) {
 	dns_rdata_t soarr = DNS_RDATA_INIT;
 	dns_rdata_soa_t soa;
 	dns_name_t primary;
+	unsigned int resoptions;
 
 	result = dns_rdataset_first(soaset);
 	if (result != ISC_R_SUCCESS)
@@ -2211,10 +2233,14 @@ process_soa(updatectx_t *uctx, dns_rdataset_t *soaset, dns_name_t *soaname) {
 		 */
 		LOCK(&uctx->lock);
 		uctx->bp4 = uctx;
+		resoptions = 0;
+		if (uctx->want_tcp)
+			resoptions |= DNS_CLIENTRESOPT_TCP;
 		result = dns_client_startresolve(uctx->client, &primary,
 						 uctx->rdclass,
 						 dns_rdatatype_a,
-						 0, uctx->client->task,
+						 resoptions,
+						 uctx->client->task,
 						 resolveaddr_done, &uctx->bp4,
 						 &uctx->restrans);
 		if (result == ISC_R_SUCCESS) {
@@ -2223,7 +2249,8 @@ process_soa(updatectx_t *uctx, dns_rdataset_t *soaset, dns_name_t *soaname) {
 							 &primary,
 							 uctx->rdclass,
 							 dns_rdatatype_aaaa,
-							 0, uctx->client->task,
+							 resoptions,
+							 uctx->client->task,
 							 resolveaddr_done,
 							 &uctx->bp6,
 							 &uctx->restrans2);
@@ -2254,7 +2281,7 @@ receive_soa(isc_task_t *task, isc_event_t *event) {
 	isc_boolean_t seencname = ISC_FALSE;
 	isc_boolean_t droplabel = ISC_FALSE;
 	dns_name_t tname;
-	unsigned int nlabels;
+	unsigned int nlabels, reqoptions;
 
 	UNUSED(task);
 
@@ -2289,9 +2316,12 @@ receive_soa(isc_task_t *task, isc_event_t *event) {
 		/* Retry SOA request without TSIG */
 		dns_message_destroy(&rcvmsg);
 		dns_message_renderreset(uctx->soaquery);
+		reqoptions = 0;
+		if (uctx->want_tcp)
+			reqoptions |= DNS_REQUESTOPT_TCP;
 		result = dns_request_createvia3(uctx->view->requestmgr,
-						uctx->soaquery, NULL, addr, 0,
-						NULL,
+						uctx->soaquery, NULL, addr,
+						reqoptions, NULL,
 						client->find_timeout * 20,
 						client->find_timeout, 3,
 						uctx->client->task,
@@ -2389,9 +2419,13 @@ receive_soa(isc_task_t *task, isc_event_t *event) {
 			UNLOCK(&uctx->lock);
 			dns_message_renderreset(soaquery);
 			dns_message_settsigkey(soaquery, NULL);
+			reqoptions = 0;
+			if (uctx->want_tcp)
+				reqoptions |= DNS_REQUESTOPT_TCP;
 			result = dns_request_createvia3(uctx->view->requestmgr,
 							soaquery, NULL,
-							uctx->currentserver, 0,
+							uctx->currentserver,
+							reqoptions,
 							uctx->tsigkey,
 							client->find_timeout *
 							20,
@@ -2422,6 +2456,7 @@ request_soa(updatectx_t *uctx) {
 	dns_message_t *soaquery = uctx->soaquery;
 	dns_name_t *name = NULL;
 	dns_rdataset_t *rdataset = NULL;
+	unsigned int reqoptions;
 
 	if (soaquery == NULL) {
 		result = dns_message_create(uctx->client->mctx,
@@ -2443,10 +2478,13 @@ request_soa(updatectx_t *uctx) {
 	dns_message_addname(soaquery, name, DNS_SECTION_QUESTION);
 	rdataset = NULL;
 	name = NULL;
+	reqoptions = 0;
+	if (uctx->want_tcp)
+		reqoptions |= DNS_REQUESTOPT_TCP;
 
 	result = dns_request_createvia3(uctx->view->requestmgr,
-					soaquery, NULL, uctx->currentserver, 0,
-					uctx->tsigkey,
+					soaquery, NULL, uctx->currentserver,
+					reqoptions, uctx->tsigkey,
 					uctx->client->find_timeout * 20,
 					uctx->client->find_timeout, 3,
 					uctx->client->task, receive_soa, uctx,
@@ -2475,7 +2513,7 @@ resolvesoa_done(isc_task_t *task, isc_event_t *event) {
 	dns_name_t *name, tname;
 	dns_rdataset_t *rdataset = NULL;
 	isc_result_t result = rev->result;
-	unsigned int nlabels;
+	unsigned int nlabels, resoptions;
 
 	UNUSED(task);
 
@@ -2516,10 +2554,14 @@ resolvesoa_done(isc_task_t *task, isc_event_t *event) {
 		dns_name_getlabelsequence(&uctx->soaqname, 1, nlabels - 1,
 					  &tname);
 		dns_name_clone(&tname, &uctx->soaqname);
+		resoptions = 0;
+		if (uctx->want_tcp)
+			resoptions |= DNS_CLIENTRESOPT_TCP;
 
 		result = dns_client_startresolve(uctx->client, &uctx->soaqname,
 						 uctx->rdclass,
-						 dns_rdatatype_soa, 0,
+						 dns_rdatatype_soa,
+						 resoptions,
 						 uctx->client->task,
 						 resolvesoa_done, uctx,
 						 &uctx->restrans);
@@ -2662,11 +2704,11 @@ dns_client_update(dns_client_t *client, dns_rdataclass_t rdclass,
 	REQUIRE(DNS_CLIENT_VALID(client));
 
 	if ((client->attributes & DNS_CLIENTATTR_OWNCTX) == 0 &&
-	    (options & DNS_CLIENTRESOPT_ALLOWRUN) == 0) {
+	    (options & DNS_CLIENTUPDOPT_ALLOWRUN) == 0) {
 		/*
 		 * If the client is run under application's control, we need
 		 * to create a new running (sub)environment for this
-		 * particular resolution.
+		 * particular update.
 		 */
 		return (ISC_R_NOTIMPLEMENTED); /* XXXTBD */
 	} else
@@ -2730,6 +2772,48 @@ dns_client_update(dns_client_t *client, dns_rdataclass_t rdclass,
 	return (result);
 }
 
+static void
+startupdate(isc_task_t *task, isc_event_t *event) {
+	updatectx_t *uctx;
+	isc_result_t result;
+	unsigned int resoptions;
+
+	REQUIRE(event != NULL);
+
+	UNUSED(task);
+
+	uctx = event->ev_arg;
+
+	if (uctx->zonename != NULL && uctx->currentserver != NULL) {
+		result = send_update(uctx);
+		if (result != ISC_R_SUCCESS)
+			goto fail;
+	} else if (uctx->currentserver != NULL) {
+		result = request_soa(uctx);
+		if (result != ISC_R_SUCCESS)
+			goto fail;
+	} else {
+		resoptions = 0;
+		if (uctx->want_tcp)
+			resoptions |= DNS_CLIENTRESOPT_TCP;
+		dns_name_clone(uctx->firstname, &uctx->soaqname);
+		result = dns_client_startresolve(uctx->client, &uctx->soaqname,
+						 uctx->rdclass,
+						 dns_rdatatype_soa, resoptions,
+						 uctx->client->task,
+						 resolvesoa_done, uctx,
+						 &uctx->restrans);
+		if (result != ISC_R_SUCCESS)
+			goto fail;
+	}
+
+	isc_event_free(&event);
+
+fail:
+	if (result != ISC_R_SUCCESS)
+		update_sendevent(uctx, result);
+}
+
 isc_result_t
 dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 		       dns_name_t *zonename, dns_namelist_t *prerequisites,
@@ -2742,10 +2826,11 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 	isc_result_t result;
 	dns_name_t *name, *newname;
 	updatectx_t *uctx;
-	isc_task_t *clone = NULL;
+	isc_task_t *tclone = NULL;
 	dns_section_t section = DNS_SECTION_UPDATE;
 	isc_sockaddr_t *server, *sa = NULL;
 	dns_tsectype_t tsectype = dns_tsectype_none;
+	isc_boolean_t want_tcp;
 
 	UNUSED(options);
 
@@ -2767,20 +2852,27 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 	if (result != ISC_R_SUCCESS)
 		return (result);
 
-	/* Create a context and prepare some resources */
+	want_tcp = ISC_TF((options & DNS_CLIENTUPDOPT_TCP) != 0);
+
+	/*
+	 * Create a context and prepare some resources.
+	 */
+
 	uctx = isc_mem_get(client->mctx, sizeof(*uctx));
 	if (uctx == NULL) {
 		dns_view_detach(&view);
 		return (ISC_R_NOMEMORY);
 	}
+
 	result = isc_mutex_init(&uctx->lock);
 	if (result != ISC_R_SUCCESS) {
 		dns_view_detach(&view);
 		isc_mem_put(client->mctx, uctx, sizeof(*uctx));
 		return (ISC_R_NOMEMORY);
 	}
-	clone = NULL;
-	isc_task_attach(task, &clone);
+
+	tclone = NULL;
+	isc_task_attach(task, &tclone);
 	uctx->client = client;
 	ISC_LINK_INIT(uctx, link);
 	uctx->state = dns_clientupdatestate_prepare;
@@ -2799,6 +2891,7 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 	uctx->tsigkey = NULL;
 	uctx->sig0key = NULL;
 	uctx->zonename = NULL;
+	uctx->want_tcp = want_tcp;
 	dns_name_init(&uctx->soaqname, NULL);
 	ISC_LIST_INIT(uctx->servers);
 	uctx->nservers = 0;
@@ -2807,7 +2900,7 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 	if (tsec != NULL)
 		dns_tsec_getkey(tsec, &uctx->tsigkey);
 	uctx->event = (dns_clientupdateevent_t *)
-		isc_event_allocate(client->mctx, clone, DNS_EVENT_UPDATEDONE,
+		isc_event_allocate(client->mctx, tclone, DNS_EVENT_UPDATEDONE,
 				   action, arg, sizeof(*uctx->event));
 	if (uctx->event == NULL)
 		goto fail;
@@ -2879,28 +2972,12 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 	ISC_LIST_APPEND(client->updatectxs, uctx, link);
 	UNLOCK(&client->lock);
 
-	if (uctx->zonename != NULL && uctx->currentserver != NULL) {
-		result = send_update(uctx);
-		if (result != ISC_R_SUCCESS)
-			goto fail;
-	} else if (uctx->currentserver != NULL) {
-		result = request_soa(uctx);
-		if (result != ISC_R_SUCCESS)
-			goto fail;
-	} else {
-		dns_name_clone(uctx->firstname, &uctx->soaqname);
-		result = dns_client_startresolve(uctx->client, &uctx->soaqname,
-						 uctx->rdclass,
-						 dns_rdatatype_soa, 0,
-						 client->task, resolvesoa_done,
-						 uctx, &uctx->restrans);
-		if (result != ISC_R_SUCCESS)
-			goto fail;
-	}
-
 	*transp = (dns_clientupdatetrans_t *)uctx;
-
-	return (ISC_R_SUCCESS);
+	result = isc_app_ctxonrun(client->actx, client->mctx, client->task,
+				  startupdate, uctx);
+	if (result == ISC_R_SUCCESS)
+		return (result);
+	*transp = NULL;
 
  fail:
 	if (ISC_LINK_LINKED(uctx, link)) {
@@ -2918,7 +2995,7 @@ dns_client_startupdate(dns_client_t *client, dns_rdataclass_t rdclass,
 		isc_event_free(ISC_EVENT_PTR(&uctx->event));
 	if (uctx->tsigkey != NULL)
 		dns_tsigkey_detach(&uctx->tsigkey);
-	isc_task_detach(&clone);
+	isc_task_detach(&tclone);
 	DESTROYLOCK(&uctx->lock);
 	uctx->magic = 0;
 	isc_mem_put(client->mctx, uctx, sizeof(*uctx));
