@@ -1,29 +1,24 @@
 /*
- * Copyright (C) 2014  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2014-2017  Internet Systems Consortium, Inc. ("ISC")
  *
- * Permission to use, copy, modify, and/or distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
- *
- * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
- * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
- * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
- * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
- * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
- * PERFORMANCE OF THIS SOFTWARE.
+ * This Source Code Form is subject to the terms of the Mozilla Public
+ * License, v. 2.0. If a copy of the MPL was not distributed with this
+ * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-/* $Id$ */
-
 #ifdef PKCS11CRYPTO
 
 #include <config.h>
 
+#include <pk11/site.h>
+
+#ifndef PK11_DSA_DISABLE
+
 #include <string.h>
 
 #include <isc/entropy.h>
 #include <isc/mem.h>
+#include <isc/safe.h>
 #include <isc/sha1.h>
 #include <isc/util.h>
 
@@ -100,16 +95,20 @@ pkcs11dsa_createctx_sign(dst_key_t *key, dst_context_t *dctx) {
 	isc_result_t ret;
 	unsigned int i;
 
+	REQUIRE(key != NULL);
+	dsa = key->keydata.pkey;
+	REQUIRE(dsa != NULL);
+
 	pk11_ctx = (pk11_context_t *) isc_mem_get(dctx->mctx,
 						  sizeof(*pk11_ctx));
 	if (pk11_ctx == NULL)
 		return (ISC_R_NOMEMORY);
 	ret = pk11_get_session(pk11_ctx, OP_DSA, ISC_TRUE, ISC_FALSE,
-			       ISC_FALSE, NULL, pk11_get_best_token(OP_DSA));
+			       dsa->reqlogon, NULL,
+			       pk11_get_best_token(OP_DSA));
 	if (ret != ISC_R_SUCCESS)
 		goto err;
 
-	dsa = key->keydata.pkey;
 	if (dsa->ontoken && (dsa->object != CK_INVALID_HANDLE)) {
 		pk11_ctx->ontoken = dsa->ontoken;
 		pk11_ctx->object = dsa->object;
@@ -230,16 +229,18 @@ pkcs11dsa_createctx_verify(dst_key_t *key, dst_context_t *dctx) {
 	isc_result_t ret;
 	unsigned int i;
 
+	dsa = key->keydata.pkey;
+	REQUIRE(dsa != NULL);
 	pk11_ctx = (pk11_context_t *) isc_mem_get(dctx->mctx,
 						  sizeof(*pk11_ctx));
 	if (pk11_ctx == NULL)
 		return (ISC_R_NOMEMORY);
 	ret = pk11_get_session(pk11_ctx, OP_DSA, ISC_TRUE, ISC_FALSE,
-			       ISC_FALSE, NULL, pk11_get_best_token(OP_DSA));
+			       dsa->reqlogon, NULL,
+			       pk11_get_best_token(OP_DSA));
 	if (ret != ISC_R_SUCCESS)
 		goto err;
 
-	dsa = key->keydata.pkey;
 	if (dsa->ontoken && (dsa->object != CK_INVALID_HANDLE)) {
 		pk11_ctx->ontoken = dsa->ontoken;
 		pk11_ctx->object = dsa->object;
@@ -388,6 +389,7 @@ pkcs11dsa_sign(dst_context_t *dctx, isc_buffer_t *sig) {
 	isc_region_t r;
 	pk11_context_t *pk11_ctx = dctx->ctxdata.pk11_ctx;
 	isc_result_t ret = ISC_R_SUCCESS;
+	unsigned int klen;
 
 	isc_buffer_availableregion(sig, &r);
 	if (r.length < ISC_SHA1_DIGESTLENGTH * 2 + 1)
@@ -399,7 +401,10 @@ pkcs11dsa_sign(dst_context_t *dctx, isc_buffer_t *sig) {
 	if (siglen != ISC_SHA1_DIGESTLENGTH * 2)
 		return (DST_R_SIGNFAILURE);
 
-	*r.base = (dctx->key->key_size - 512)/64;
+	klen = (dctx->key->key_size - 512)/64;
+	if (klen > 255)
+		return (ISC_R_FAILURE);
+	*r.base = klen;
 	isc_buffer_add(sig, ISC_SHA1_DIGESTLENGTH * 2 + 1);
 
     err:
@@ -439,7 +444,8 @@ pkcs11dsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 		return (ISC_TRUE);
 	else if ((attr1 == NULL) || (attr2 == NULL) ||
 		 (attr1->ulValueLen != attr2->ulValueLen) ||
-		 memcmp(attr1->pValue, attr2->pValue, attr1->ulValueLen))
+		 !isc_safe_memequal(attr1->pValue, attr2->pValue,
+				    attr1->ulValueLen))
 		return (ISC_FALSE);
 
 	attr1 = pk11_attribute_bytype(dsa1, CKA_SUBPRIME);
@@ -448,7 +454,8 @@ pkcs11dsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 		return (ISC_TRUE);
 	else if ((attr1 == NULL) || (attr2 == NULL) ||
 		 (attr1->ulValueLen != attr2->ulValueLen) ||
-		 memcmp(attr1->pValue, attr2->pValue, attr1->ulValueLen))
+		 !isc_safe_memequal(attr1->pValue, attr2->pValue,
+				    attr1->ulValueLen))
 		return (ISC_FALSE);
 
 	attr1 = pk11_attribute_bytype(dsa1, CKA_BASE);
@@ -457,7 +464,8 @@ pkcs11dsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 		return (ISC_TRUE);
 	else if ((attr1 == NULL) || (attr2 == NULL) ||
 		 (attr1->ulValueLen != attr2->ulValueLen) ||
-		 memcmp(attr1->pValue, attr2->pValue, attr1->ulValueLen))
+		 !isc_safe_memequal(attr1->pValue, attr2->pValue,
+				    attr1->ulValueLen))
 		return (ISC_FALSE);
 
 	attr1 = pk11_attribute_bytype(dsa1, CKA_VALUE);
@@ -466,7 +474,8 @@ pkcs11dsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 		return (ISC_TRUE);
 	else if ((attr1 == NULL) || (attr2 == NULL) ||
 		 (attr1->ulValueLen != attr2->ulValueLen) ||
-		 memcmp(attr1->pValue, attr2->pValue, attr1->ulValueLen))
+		 !isc_safe_memequal(attr1->pValue, attr2->pValue,
+				    attr1->ulValueLen))
 		return (ISC_FALSE);
 
 	attr1 = pk11_attribute_bytype(dsa1, CKA_VALUE2);
@@ -474,7 +483,8 @@ pkcs11dsa_compare(const dst_key_t *key1, const dst_key_t *key2) {
 	if (((attr1 != NULL) || (attr2 != NULL)) &&
 	    ((attr1 == NULL) || (attr2 == NULL) ||
 	     (attr1->ulValueLen != attr2->ulValueLen) ||
-	     memcmp(attr1->pValue, attr2->pValue, attr1->ulValueLen)))
+	     !isc_safe_memequal(attr1->pValue, attr2->pValue,
+				attr1->ulValueLen)))
 		return (ISC_FALSE);
 
 	if (!dsa1->ontoken && !dsa2->ontoken)
@@ -744,23 +754,25 @@ pkcs11dsa_todns(const dst_key_t *key, isc_buffer_t *data) {
 		return (ISC_R_NOSPACE);
 
 	memset(r.base, 0, dnslen);
-	*r.base++ = t;
+	*r.base = t;
+	isc_region_consume(&r, 1);
+
 	cp = (CK_BYTE *) subprime->pValue;
 	memmove(r.base + ISC_SHA1_DIGESTLENGTH - subprime->ulValueLen,
 		cp, subprime->ulValueLen);
-	r.base += ISC_SHA1_DIGESTLENGTH;
+	isc_region_consume(&r, ISC_SHA1_DIGESTLENGTH);
 	cp = (CK_BYTE *) prime->pValue;
 	memmove(r.base + key->key_size/8 - prime->ulValueLen,
 		cp, prime->ulValueLen);
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 	cp = (CK_BYTE *) base->pValue;
 	memmove(r.base + key->key_size/8 - base->ulValueLen,
 		cp, base->ulValueLen);
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 	cp = (CK_BYTE *) pub_key->pValue;
 	memmove(r.base + key->key_size/8 - pub_key->ulValueLen,
 		cp, pub_key->ulValueLen);
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 
 	isc_buffer_add(data, dnslen);
 
@@ -784,7 +796,8 @@ pkcs11dsa_fromdns(dst_key_t *key, isc_buffer_t *data) {
 		return (ISC_R_NOMEMORY);
 	memset(dsa, 0, sizeof(*dsa));
 
-	t = (unsigned int) *r.base++;
+	t = (unsigned int) *r.base;
+	isc_region_consume(&r, 1);
 	if (t > 8) {
 		memset(dsa, 0, sizeof(*dsa));
 		isc_mem_put(key->mctx, dsa, sizeof(*dsa));
@@ -792,23 +805,23 @@ pkcs11dsa_fromdns(dst_key_t *key, isc_buffer_t *data) {
 	}
 	p_bytes = 64 + 8 * t;
 
-	if (r.length < 1 + ISC_SHA1_DIGESTLENGTH + 3 * p_bytes) {
+	if (r.length < ISC_SHA1_DIGESTLENGTH + 3 * p_bytes) {
 		memset(dsa, 0, sizeof(*dsa));
 		isc_mem_put(key->mctx, dsa, sizeof(*dsa));
 		return (DST_R_INVALIDPUBLICKEY);
 	}
 
 	subprime = r.base;
-	r.base += ISC_SHA1_DIGESTLENGTH;
+	isc_region_consume(&r, ISC_SHA1_DIGESTLENGTH);
 
 	prime = r.base;
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 
 	base = r.base;
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 
 	pub_key = r.base;
-	r.base += p_bytes;
+	isc_region_consume(&r, p_bytes);
 
 	key->key_size = p_bytes * 8;
 
@@ -1095,6 +1108,7 @@ dst__pkcs11dsa_init(dst_func_t **funcp) {
 		*funcp = &pkcs11dsa_functions;
 	return (ISC_R_SUCCESS);
 }
+#endif /* !PK11_DSA_DISABLE */
 
 #else /* PKCS11CRYPTO */
 
